%{
/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lkc.h"

#define printd(mask, fmt...) if (cdebug & (mask)) printf(fmt)

#define PRINTD		0x0001
#define DEBUG_PARSE	0x0002

int cdebug = PRINTD;

extern int zconflex(void);
static void zconferror(const char *err);

struct symbol *symbol_hash[257];

#define YYERROR_VERBOSE
%}

%union
{
	char *string;
	struct symbol *symbol;
	struct expr *expr;
	struct statement *stmt;
	struct menu *menu;
}

%token T_MAINMENU
%token T_MENU
%token T_ENDMENU
%token T_SOURCE
%token T_CHOICE
%token T_ENDCHOICE
%token T_COMMENT
%token T_CONFIG
%token T_HELP
%token <string> T_HELPTEXT
%token T_IF 304
%token T_ENDIF
%token T_DEPENDS
%token T_OPTIONAL
%token T_PROMPT 306
%token T_DEFAULT 307
%token T_TRISTATE, T_BOOLEAN, T_INT, T_HEX, T_STRING
%token <string> T_WORD 300, T_STRING
%token T_UNEQUAL 301

%left '|'
%left '&'
%left '=' T_UNEQUAL
%nonassoc '!'

%type <string> prompt
%type <symbol> symbol
%type <expr> expr, if_expr
%type <stmt> statement, menu_stmt, if_stmt, choice_stmt, source_stmt

%%
input:  block
;

block:	  /* empty */
	| block config_stmt
	| block choice_stmt
	| block menu_stmt
	| block if_stmt
	| block source_stmt
	| block statement
;

config_stmt: T_CONFIG T_WORD '\n'
{
	struct symbol *sym = lookup_symbol($2, 0);
	printd(DEBUG_PARSE, "%s:%d:config %s\n", zconf_curname(), zconf_lineno(), $2);
	sym->flags |= SYMBOL_OPTIONAL;
	menu_add_entry(sym);
}
	config_option_list
{
	printd(DEBUG_PARSE, "%s:%d:endconfig\n", zconf_curname(), zconf_lineno());
	menu_end_entry();
};

config_option_list:
	  /* empty */
	| config_option_list config_option
{
};

config_option: T_DEPENDS expr '\n'
{
	menu_add_dep($2);
	printd(DEBUG_PARSE, "%s:%d:depends\n", zconf_curname(), zconf_lineno());
};

config_option: T_TRISTATE '\n'
{
	menu_set_type(SYMBOL_TRISTATE);
	printd(DEBUG_PARSE, "%s:%d:tristate\n", zconf_curname(), zconf_lineno());
};

config_option: T_TRISTATE prompt if_expr '\n'
{
	menu_set_type(SYMBOL_TRISTATE);
	menu_add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:tristate\n", zconf_curname(), zconf_lineno());
};

config_option: T_BOOLEAN '\n'
{
	menu_set_type(SYMBOL_BOOLEAN);
	printd(DEBUG_PARSE, "%s:%d:boolean\n", zconf_curname(), zconf_lineno());
};

config_option: T_BOOLEAN prompt if_expr '\n'
{
	menu_set_type(SYMBOL_BOOLEAN);
	menu_add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:boolean\n", zconf_curname(), zconf_lineno());
};

config_option: T_INT '\n'
{
	menu_set_type(SYMBOL_INT);
	printd(DEBUG_PARSE, "%s:%d:int\n", zconf_curname(), zconf_lineno());
};

config_option: T_INT prompt if_expr '\n'
{
	menu_set_type(SYMBOL_INT);
	menu_add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:int\n", zconf_curname(), zconf_lineno());
};

config_option: T_HEX '\n'
{
	menu_set_type(SYMBOL_HEX);
	printd(DEBUG_PARSE, "%s:%d:hex\n", zconf_curname(), zconf_lineno());
};

config_option: T_HEX prompt if_expr '\n'
{
	menu_set_type(SYMBOL_HEX);
	menu_add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:hex\n", zconf_curname(), zconf_lineno());
};

config_option: T_STRING '\n'
{
	menu_set_type(SYMBOL_STRING);
	printd(DEBUG_PARSE, "%s:%d:string\n", zconf_curname(), zconf_lineno());
};

config_option: T_STRING prompt if_expr '\n'
{
	menu_set_type(SYMBOL_STRING);
	menu_add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:string\n", zconf_curname(), zconf_lineno());
};

config_option: T_PROMPT prompt if_expr '\n'
{
	menu_add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:prompt\n", zconf_curname(), zconf_lineno());
};

config_option: T_DEFAULT symbol if_expr '\n'
{
	menu_add_stmt(T_DEFAULT, NULL, $2, $3);
	printd(DEBUG_PARSE, "%s:%d:default\n", zconf_curname(), zconf_lineno());
};

config_option: T_HELP '\n'
{
	printd(DEBUG_PARSE, "%s:%d:help\n", zconf_curname(), zconf_lineno());
	zconf_starthelp();
}
	T_HELPTEXT
{
	current_entry->sym->help = $4;
};

choice_stmt: T_CHOICE '\n'
{
	struct symbol *sym = lookup_symbol(NULL, 0);
	sym->flags |= SYMBOL_CHOICE;
	menu_add_entry(sym);
	menu_add_stmt('^', NULL, NULL, NULL);
	printd(DEBUG_PARSE, "%s:%d:choice\n", zconf_curname(), zconf_lineno());
}
	choice_option_list
{
	menu_end_entry();
	menu_add_menu();
}
	choice_block T_ENDCHOICE '\n'
{
	menu_end_menu();
	printd(DEBUG_PARSE, "%s:%d:endchoice\n", zconf_curname(), zconf_lineno());
};

choice_option_list:
	  /* empty */
	| choice_option_list choice_option
;

choice_option: T_DEPENDS expr '\n'
{
	menu_add_dep($2);
	printd(DEBUG_PARSE, "%s:%d:depends\n", zconf_curname(), zconf_lineno());
};

choice_option: T_PROMPT prompt '\n'
{
	menu_add_stmt(T_PROMPT, $2, NULL, NULL);
	printd(DEBUG_PARSE, "%s:%d:prompt\n", zconf_curname(), zconf_lineno());
};

choice_option: T_OPTIONAL '\n'
{
	current_entry->sym->flags |= SYMBOL_OPTIONAL;
	printd(DEBUG_PARSE, "%s:%d:optional\n", zconf_curname(), zconf_lineno());
};

choice_option: T_DEFAULT symbol '\n'
{
	menu_add_stmt(T_DEFAULT, NULL, $2, NULL);
	//current_choice->stmt->def = $2;
	printd(DEBUG_PARSE, "%s:%d:default\n", zconf_curname(), zconf_lineno());
};

choice_option: T_HELP '\n'
{
	printd(DEBUG_PARSE, "%s:%d:help\n", zconf_curname(), zconf_lineno());
	zconf_starthelp();
}
	T_HELPTEXT
{
	current_entry->sym->help = $4;
};

choice_block:	  /* empty */
		| choice_block config_stmt
		| choice_block source_stmt
		| choice_block '\n'
;

if_stmt: T_IF expr '\n'
{
	printd(DEBUG_PARSE, "%s:%d:if\n", zconf_curname(), zconf_lineno());
	menu_add_entry(NULL);
	//current_entry->prompt = menu_add_stmt(T_IF, NULL, NULL, $2);
	menu_add_dep($2);
	menu_end_entry();
	menu_add_menu();
}
	menu_block T_ENDIF '\n'
{
	menu_end_menu();
	printd(DEBUG_PARSE, "%s:%d:endif\n", zconf_curname(), zconf_lineno());
};

menu_stmt: T_MENU prompt '\n'
{
	printd(DEBUG_PARSE, "%s:%d:menu\n", zconf_curname(), zconf_lineno());
	menu_add_entry(NULL);
	menu_add_stmt(T_MENU, $2, NULL, NULL);
}
	stmt_dep_opt
{
	menu_end_entry();
	menu_add_menu();
}
	menu_block T_ENDMENU '\n'
{
	menu_end_menu();
	printd(DEBUG_PARSE, "%s:%d:endmenu\n", zconf_curname(), zconf_lineno());
};

menu_block:	  /* empty */
		| menu_block menu_stmt
		| menu_block if_stmt
		| menu_block choice_stmt
		| menu_block config_stmt
		| menu_block source_stmt
		| menu_block statement
;

statement: '\n'
{
	$$ = NULL;
};

statement: T_MAINMENU prompt '\n'
{
	printd(DEBUG_PARSE, "%s:%d:mainmenu\n", zconf_curname(), zconf_lineno());
	//$$ = create_stmt(T_MAINMENU);
	//$$->prompt = $2;
};

source_stmt: T_SOURCE prompt '\n'
{
	//$$ = create_stmt(T_SOURCE);
	//$$->prompt = $2;
	zconf_nextfile($2);
	printd(DEBUG_PARSE, "%s:%d:source %s\n", zconf_curname(), zconf_lineno(), $2);
};

statement: T_COMMENT prompt '\n'
{
	menu_add_entry(NULL);
	menu_add_stmt(T_COMMENT, $2, NULL, NULL);
	printd(DEBUG_PARSE, "%s:%d:comment\n", zconf_curname(), zconf_lineno());
}
	stmt_dep_opt
{
	menu_end_entry();
};

stmt_dep_opt:
	  /* empty */
	| T_DEPENDS expr '\n'
{
	menu_add_dep($2);
	printd(DEBUG_PARSE, "%s:%d:depends\n", zconf_curname(), zconf_lineno());
};

prompt:	  T_WORD
	| T_STRING
;

if_expr:  /* empty */			{ $$ = NULL; }
	| T_IF expr			{ $$ = $2; }
;

expr:	  symbol			{ $$ = expr_alloc_symbol($1); }
	| symbol '=' symbol		{ $$ = expr_alloc_comp('=', $1, $3); }
	| symbol T_UNEQUAL symbol	{ $$ = expr_alloc_comp(T_UNEQUAL, $1, $3); }
	| '(' expr ')'			{ $$ = $2; }
	| '!' expr			{ $$ = expr_alloc_one('!', $2); }
	| expr '|' expr			{ $$ = expr_alloc_two('|', $1, $3); }
	| expr '&' expr			{ $$ = expr_alloc_two('&', $1, $3); }
;

symbol:	  T_WORD	{ $$ = lookup_symbol($1, 0); free($1); }
	| T_STRING	{ $$ = lookup_symbol($1, 1); free($1); }
;

%%

void conf_parse(const char *name)
{
	zconf_initscan(name);

	menu_init();
	rootmenu.prompt = create_stmt(T_MENU);

	zconfparse();
	menu_finalize(&rootmenu);
	modules_sym = lookup_symbol("MODULES", 0);
}

static void zconferror(const char *err)
{
	printf("%s:%d:%s\n", zconf_curname(), zconf_lineno(), err);
	exit(1);
}

struct symbol *lookup_symbol(char *name, int isconst)
{
	struct symbol *symbol;
	char *ptr;
	int hash = 0;

	//printf("lookup: %s -> ", name);
	if (name) {
		if (name[0] && !name[1]) {
			switch (name[0]) {
			case 'y': return &symbol_yes;
			case 'm': return &symbol_mod;
			case 'n': return &symbol_no;
			}
		}
		for (ptr = name; *ptr; ptr++)
			hash += *ptr;
		hash &= 0xff;

		for (symbol = symbol_hash[hash]; symbol; symbol = symbol->next) {
			if (!strcmp(symbol->name, name)) {
				if ((isconst && symbol->flags & SYMBOL_CONST) ||
				    (!isconst && !(symbol->flags & SYMBOL_CONST))) {
					//printf("h:%p\n", symbol);
					return symbol;
				}
			}
		}
		name = strdup(name);
	} else
		hash = 256;

	symbol = malloc(sizeof(*symbol));
	memset(symbol, 0, sizeof(*symbol));
	symbol->name = name;
	symbol->type = SYMBOL_UNKNOWN;
	symbol->flags = SYMBOL_NEW;
	if (isconst)
		symbol->flags |= SYMBOL_CONST;

	symbol->next = symbol_hash[hash];
	symbol_hash[hash] = symbol;

	//printf("n:%p\n", symbol);
	return symbol;
}

void print_quoted_string(FILE *out, const char *str)
{
	const char *p;
	int len;

	putc('"', out);
	while ((p = strchr(str, '"'))) {
		len = p - str;
		if (len)
			fprintf(out, "%.*s", len, str);
		fputs("\\\"", out);
		str = p + 1;
	}
	fputs(str, out);
	putc('"', out);
}

void print_symbol(FILE *out, struct menu *menu)
{
	struct symbol *sym = menu->sym;
	struct statement *stmt;

	//sym->flags |= SYMBOL_PRINTED;

	if (sym_is_choice(sym))
		fprintf(out, "choice\n");
	else
		fprintf(out, "config %s\n", sym->name);
	switch (sym->type) {
	case SYMBOL_BOOLEAN:
		fputs("  boolean\n", out);
		break;
	case SYMBOL_TRISTATE:
		fputs("  tristate\n", out);
		break;
	case SYMBOL_STRING:
		fputs("  string\n", out);
		break;
	case SYMBOL_INT:
		fputs("  integer\n", out);
		break;
	case SYMBOL_HEX:
		fputs("  hex\n", out);
		break;
	}
#if 0
	if (!expr_is_yes(sym->dep)) {
		fputs("  depends ", out);
		fprint_expr(out, sym->dep, 0);
		fputc('\n', out);
	}
#endif
	for (stmt = sym->stmt; stmt; stmt = stmt->next) {
		if (stmt->menu != menu)
			continue;
		switch (stmt->token) {
		case T_PROMPT:
			fputs("  prompt ", out);
			print_quoted_string(out, stmt->prompt);
			if (stmt->def) {
				fputc(' ', out);
				if (stmt->def->flags & SYMBOL_CONST)
					print_quoted_string(out, stmt->def->name);
				else
					fputs(stmt->def->name, out);
			}
			if (!expr_is_yes(E_EXPR(stmt->visible))) {
				fputs(" if ", out);
				fprint_expr(out, E_EXPR(stmt->visible), 0);
			}
			fputc('\n', out);
			break;
		case T_DEFAULT:
			fputs( "  default ", out);
			print_quoted_string(out, stmt->def->name);
			if (!expr_is_yes(E_EXPR(stmt->visible))) {
				fputs(" if ", out);
				fprint_expr(out, E_EXPR(stmt->visible), 0);
			}
			fputc('\n', out);
			break;
		case '^':
			fputs("  #choice value\n", out);
			break;
		default:
			fprintf(out, "  unknown stmt %d!\n", stmt->token);
			break;
		}
	}
	if (sym->help) {
		int len = strlen(sym->help);
		while (sym->help[--len] == '\n')
			sym->help[len] = 0;
		fprintf(out, "  help\n%s\n", sym->help);
	}
	fputc('\n', out);
}

void zconfdump(FILE *out)
{
	//struct file *file;
	struct statement *stmt;
	struct symbol *sym;
	struct menu *menu;

	print_type = 0;

	menu = rootmenu.list;
	while (menu) {
		if ((sym = menu->sym))
			print_symbol(out, menu);
		else if ((stmt = menu->prompt)) {
			switch (stmt->token) {
			case T_MAINMENU:
				fputs("\nmainmenu ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n", out);
				break;
			case T_COMMENT:
				fputs("\ncomment ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n", out);
				break;
			case T_MENU:
				fputs("\nmenu ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n", out);
				break;
			case T_ENDMENU:
				fputs("\nendmenu\n", out);
				break;
			case T_SOURCE:
				fputs("\nsource ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n", out);
				break;
			case T_IF:
				fputs("\nif\n", out);
			}
			if (!expr_is_yes(E_EXPR(stmt->visible))) {
				fputs("  depends ", out);
				fprint_expr(out, E_EXPR(stmt->visible), 0);
				fputc('\n', out);
			}
			fputs("\n", out);
		}

		if (menu->list)
			menu = menu->list;
		else if (menu->next)
			menu = menu->next;
		else while ((menu = menu->parent)) {
			if (menu->prompt && menu->prompt->token == T_MENU)
				fputs("\nendmenu\n", out);
			if (menu->next) {
				menu = menu->next;
				break;
			}
		}
	}
}
